
/*	Load / Save snapshots 

***	Copyright:	1994-1996 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	01.aug.94	started work (on application.c)				KIO !
	05.dec.94	replaced 'core' save&load by '.sna'			KIO !
	06.dec.94	added '.z80' v1.45 loader					KIO !
	10.dec.94	enhanced '.z80' loader to version 2.01		KIO !
	14.jan.95	separated this file from application.c			KIO !
	22.jan.95	save '.z80' added						KIO !
	05.nov.95	paged memory model support				KIO !
    20.Oct.01   revised for CW7Pro and ANSI C compliance					hdusel@bnv-gz.de
*/

#include	<files.h>
#include	<memory.h>
#include	<string.h>

#include	"kio.h"
#include	"z80.h"
#include	"load+save.h"
#include	"application.h"
#include	"messages.h"
#include	"terminal.h"
#include	"display.h"
#include	"z80 main.h"
#include	"strings.h"
#include	"menue.h"
#include	"sound.h"

short	romID = 0;


// *****************************************************************


/* -----	Compress .z80 file image -------------------------------------------
		compression sheme:  dc.b $ed, $ed, count, char
		return 0  --> ok --> compressed
		return !0 --> ok --> compressed, but compression increased size
		For security: the compression buffer must be at least  (5/3*sourcesize)  bytes long
*/
OSErr CompressZ80    (Char *source, long sourcesize, Char *dest, long *destsize);
OSErr CompressZ80    (Char *source, long sourcesize, Char *dest, long *destsize)
{	Char *q,*z,c,n,*qe;
	q = source;
	z = dest;
	qe = source+ sourcesize;
	
	while (q<qe)
	{	c=*q++;
		if (*q!=c)						// single byte
		{	*z++ = c;
			if ((c==0xed)&&(q+2<=qe)&&(*q==*(q+1)))	// special care for compressible sequence after single 0xed !!!
				*z++=*q++;
			continue;
		}
		for (n=1; n<255 && *q==c; n++) q++;
		if (q>qe) { n -= q-qe; q=qe; };
		if ( (n>=4) || (c==0xed) )			// compress ?
		{	*z++ = 0xed;
			*z++ = 0xed;
			*z++ = n;
			*z++ = c;
		}
		else							// don't compress
		{	for (;n;n--) *z++ = c;
		}
	}
	
	*destsize = z-dest;
	return ( (*destsize<=sourcesize) ? ok : error );
}


/* -----	Decompress .z80 file image -----------------------------------------
		compression scheme:  dc.b $ed, $ed, count, char
		return 0:	ok
		return 1:	error
*/
OSErr DecompressZ80 (Char *source, long sourcesize, Char *dest, long destsize);
OSErr DecompressZ80 (Char *source, long sourcesize, Char *dest, long destsize)
{	Char *q,*z,c,n;

	q=source;	source+=sourcesize;
	z=dest;	dest+=destsize;
	
	while ( z<dest && q<source ) 
	{	c=*q++;						// read next byte
		if (c!=0xed) { *z++=c; continue; }	// single byte

		if (q==source) { *z++=c; continue; } // rare case: last source byte is 0xed

		c=*q++;						// read next byte
		if (c!=0xed) { *z++=0xed; q--; continue; };
		
		n=*q++; c=*q++;				// read count & char
		if ( (z+n)>dest ) return error;
		if (q>source) return error;
		memset (z, c, n);				// expand compressed bytes
		z+=n;
	};
	if (q==source&&z==dest) return ok;

	if (z==dest&&q+4<=source&&destsize==0xC000)	// version 1.45 file with end marker ?
		if (*q++==0) if (*q++==0xed) if (*q++==0xed) if (*q++==0) return ok;

	return error;
}


/* -----	save one page in .z80 v2.01 format ---------------------------------------------------
		5.nov.95:	now takes pointer to page as argument
*/
OSErr save_block_z80 ( short fileID, Char ID, Char* page ); 
OSErr save_block_z80 ( short fileID, Char ID, Char* page )
{	Handle	scratch;
	long		destsize;
	OSErr 	err;
	Char		*p;
	
	scratch = TempNewHandle ( 0x4000*5/3+4, &err );
	if (err||!scratch) { PrintAlert(CStr(localized("No space for compression buffer","Kein Platz für den Kompressionspuffer"))); return err; };
	HLock(scratch);
	
	p = (Char*)*scratch;
	CompressZ80 ( page, 0x4000, p+3, &destsize);
	*p++ = destsize;
	*p++ = destsize>>8;
	*p++ = ID;
	
	err = Buffer2File ( *scratch, fileID, destsize+3 );
 	DisposHandle(scratch);
	if (err) PrintNumberAlert(CStr(localized("Error while writing RAM image:","Fehler beim Schreiben des Speicherauszuges:")),err);
 	return err;
 }


/* -----	Save Core as .sna file ------------------------------------------------------------
		note: .sna files were originally saved from an NMI routine. 
		therefore the PC is stored on the Z80 stack and only IFF1 needs to be saved
*/
OSErr SaveSNA ( short fileID )	
{	OSErr 	err;
	struct { Char i,l2,h2,e2,d2,c2,b2,f2,a2,l,h,e,d,c,b,yl,yh,xl,xh,iff,r,f,a,spl,sph,im,brdr; } head;
	
	PrintMsg(CStr("Save .sna snapshot file")); 

	if (z80_hardware!=zx48) PrintAlert ( CStr(localized(
		"Warning: The saved .sna snapshot contains only the 48 kbyte RAM of the ZX Spectrum 48k",
		"Achtung: Der gesicherte .sna Schnappschuß enthält nur die Daten für einen ZX Spectrum 48k"  )));

	Poke(--SP,PCH);		// push retn-address
	Poke(--SP,PCL);
	
	head.im	= IM;	head.iff	= IFF1 ? 0x04 : 0x00;
	head.i	= RI;		head.r	= RR;
	head.a	= RA;	head.a2	= RA2;
	head.f	= RF;	head.f2	= RF2;
	head.b	= RB;	head.c	= RC;
	head.d	= RD;	head.e	= RE;
	head.h	= RH;	head.l	= RL;
	head.xh	= XH;	head.xl	= XL;
	head.yh	= YH;	head.yl	= YL;
	head.b2	= BC2>>8;	head.c2	= BC2;
	head.d2	= DE2>>8;	head.e2	= DE2;
	head.h2	= HL2>>8;	head.l2	= HL2;
	head.sph	= SPH;	head.spl	= SPL;
	head.brdr	= bordercolor;

	err = Buffer2File ( &head, fileID, 27 );
	if (err) { PrintNumberAlert(CStr(localized("Error while writing registers:","Fehler beim Schreiben der Register:")),err); return err; }

	err = Core2File ( 0x4000, fileID, 0xc000 );	
	if (err) PrintNumberAlert(CStr(localized("Error while writing RAM image:","Fehler beim Schreiben des Speicherauszuges:")),err);
	return err;
}


OSErr LoadSNA ( short fileID )
{	OSErr 	err;
	struct { Char i,l2,h2,e2,d2,c2,b2,f2,a2,l,h,e,d,c,b,yl,yh,xl,xh,iff,r,f,a,spl,sph,im,brdr; } head;
	
	PrintMsg(CStr("Load .sna snapshot file")); 

	if (z80_hardware != zx48) HandleOptionsMenu ( i48k );
	else HandleReset(!romID);
	HandleFileMenu(iResume);
	
	err = File2Buffer(fileID,&head,27);		
	if (err) { PrintNumberAlert(CStr(localized("Error while reading registers:","Fehler beim Lesen der Register:")),err); return err; }

	err = File2Core(fileID,0x04000,0x0c000);
	if (err) { PrintNumberAlert(CStr(localized("Error reading RAM image:","Fehler beim Lesen des Speicherauszuges:")),err); return err; }

	RI 	= head.i;		RR 	= head.r;
	RA 	= head.a;		RA2	= head.a2;		
	RF	= head.f;		RF2	= head.f2;
	RB	= head.b;		RC	= head.c;
	RD	= head.d;		RE	= head.e;
	RH	= head.h;		RL	= head.l;
	XH	= head.xh;		XL	= head.xl;	
	YH	= head.yh;		YL	= head.yl;	
	SPH	= head.sph;	SPL	= head.spl;	
	BC2	= head.b2*256+head.c2;	
	DE2	= head.d2*256+head.e2;
	HL2	= head.h2*256+head.l2;
	IM	= head.im;		
	IFF1=IFF2	= (head.iff&0x04) ? enabled : disabled;
	bordercolor = head.brdr;
	PCL	= Peek(SP++);
	PCH	= Peek(SP++);	

	if (SP<=0x4000) { PrintAlert(CStr(localized("Illegal stack pointer","Stackpointer nicht gültig"))); return error; }
	if (IM>2) { PrintAlert(CStr(localized("Illegal irpt mode","Interruptmodus nicht gültig"))); return error; };
	
	return ok;
}


// -----	save .z80 file; version 2.01 --------------------------------------------------
OSErr SaveZ80 ( short fileID )
{	OSErr err;
	int i;

	struct {	Char 					// Z80 Version 1.45 Header
			a,f,c,b,l,h,pcl,pch,spl,sph,
			i,r,data,e,d,c2,b2,e2,d2,l2,h2,a2,f2,
			yl,yh,xl,xh,iff1,iff2,im,
									// Z80 Version 2.01 enhanced header (if PC=0)
			h2lenl,h2lenh,				// size of header extension: 23 for version 2.01 files
			npcl,npch,					// PC (instead of pcl+pch, which is set to 0)
			model,					// 	Value:	Meaning in v2.01	Meaning in v3.0
									//	0		48k				48k
									//	1		48k + If.1			48k + If.1
									//	2		SamRam			48k + M.G.T.
									//	3		128k				SamRam
									//	4		128k + If.1		128k
									//	5		-				128k + If.1
									//	6		-				128k + M.G.T.
			port_7ffd,				// If in SamRam mode, bitwise state of 74ls259.
									// For example, bit 6=1 after an OUT 31,13 (=2*6+1)
									// If in 128 mode, contains last OUT to 7ffd (paging control)
			if1paged,					// !=0 means: interface1 rom is paged in
			rldiremu,					// Bit 0: 1 if R register emulation on
									// Bit 1: 1 if LDIR emulation on
			port_fffd,					// Last OUT to fffd (soundchip register number)
			soundreg[16];				// Contents of the sound chip registers
		} head;

	PrintMsg(CStr("Save .z80 snapshot file")); 

	memset ( &head, 0,sizeof(head) );			// head.pcl = head.pch = head.if1paged = 0;

	head.h2lenl	= 23;	head.h2lenh = 0;	// flag: version 2.01 header
	
	switch(z80_hardware)
	{	case zx48:	head.model = 0;	 break;	// ZX Spectrum 48k
		case zx128:	head.model = 3; break;	// ZX 128
		case zxplus2:	head.model = 3; break;	// ZX 128 instead of +2  ((•••preliminary•••))
	}

	head.port_fffd	= AYregnr;				// ZX 128 only: value of AY register select register
	memcpy ( head.soundreg, AYreg, 16 );		// ZX 128 only: AY registers
	head.port_7ffd	= zx128pagectrl;			// ZX 128 only: value of page control register

	head.im	= IM;
	head.i	= RI;
	head.r	= RR & 0x7F;
	head.a	= RA;
	head.a2	= RA2;
	head.f	= RF;
	head.f2	= RF2;
	head.b	= RB;		head.c	= RC;
	head.d	= RD;		head.e	= RE;
	head.h	= RH;		head.l	= RL;
	head.xh	= XH;		head.xl	= XL;
	head.yh	= YH;		head.yl	= YL;
	head.b2	= BC2>>8;		head.c2	= BC2;
	head.d2	= DE2>>8;		head.e2	= DE2;
	head.h2	= HL2>>8;		head.l2	= HL2;
	head.sph	= SPH;		head.spl	= SPL;
	head.npcl	= PCL;		head.npch	= PCH;

	head.iff1 = IFF1;
	head.iff2 = IFF2;
	
	head.data = ((RR&0x80)>>7) + bordercolor<<1;

	err = Buffer2File(&head,fileID,55);
	if (err) { PrintNumberAlert(CStr(localized("Error while writing registers:","Fehler beim Schreiben der Register:")),err); return err; }


/*	Now the compressed blocks follow. Each block has a 3 byte header:
		dc.w		length of data (without this header; low byte first)
		dc.b		page number of block
		dc.s		compressed data follows

	The pages are numbered, depending on the hardware mode, in the following way:

		Page		48 mode			128 mode			SamRam mode
		0		48K rom			rom (basic)		48K rom
		1		Interface I, Disciple or Plus D rom, according to setting
		2		-				rom (reset)		samram rom (basic)
		3		-				page 0			samram rom (monitor,..)
		4		8000-bfff			page 1			Normal 8000-bfff
		5		c000-ffff			page 2			Normal c000-ffff
		6		-				page 3			Shadow 8000-bfff
		7		-				page 4			Shadow c000-ffff
		8		4000-7fff			page 5			4000-7fff
		9		-				page 6			-
		10		-				page 7			-
		11		Multiface rom		Multiface rom		-

	In 48K mode, pages 4,5 and 8 are saved.
	In SamRam mode, pages 4 to 8 are saved.  
	In '128 mode, all pages from 3 to 10 are saved.

	The 128 has a memory map like:   Rom [switchable];   Ram 5;   Ram 2;   Ram [switchable]
	The paging is all controlled by OUT commands to 7FFDh (32765). This port is 
	write-only so a copy of the last byte sent to it is stored at 5B5Ch (23388).
	
	The byte to output is:
		Bits 0-2:	Ram for top 16k:	0 is the usual bank;   7 contains some scratchpads used by the 128k editor.
		Bit 3:	Screen select:		0 => Use the normal screen in bank 5;	1 => Use the shadow screen in bank 7.
		Bit 4:	ROM for lower 16k:	0 => 128k editor;					1 => 48k BASIC
		Bit 5:	Lock. If set, disables all further paging.
*/

	if (z80_hardware==zx48)
	{			err = save_block_z80 ( fileID, 4, RdPtr(0x8000) );
		if (!err)	err = save_block_z80 ( fileID, 5, RdPtr(0xC000) );
		if (!err)	err = save_block_z80 ( fileID, 8, RdPtr(0x4000) );
	}

	if (z80_hardware==zx128||z80_hardware==zxplus2)
	{	for (i=0;i<=7&&!err;i++)   err = save_block_z80 ( fileID, i+3, ram[i] );
	}

	if (err) PrintNumberAlert(CStr(localized("Error while writing core:","Fehler beim Schreiben der Speicherabzuges:")),err);

	return err;
}


// ----	load .z80 file; version 1.45, 2.01, or version 3.0
OSErr LoadZ80 ( short fileID )
{	OSErr 	err;

	struct {	Char 
			// ----- Z80 Version 1.45 Header
			a,f,c,b,l,h,pcl,pch,spl,sph,
			i,r,data,e,d,c2,b2,e2,d2,l2,h2,a2,f2,
			yl,yh,xl,xh,iff1,iff2,im,
			// ----- Z80 version 2.01 enhanced header (if PC=0)
			h2lenl,h2lenh,				// size of header extension: 23 for version 2.01 files
			npcl,npch,					// PC (instead of pcl+pch, which is set to 0)
			model,					// 	Value:	Meaning in v2.01	Meaning in v3.0
									//	0		48k				48k
									//	1		48k + If.1			48k + If.1
									//	2		SamRam			48k + M.G.T.
									//	3		128k				SamRam
									//	4		128k + If.1		128k
									//	5		-				128k + If.1
									//	6		-				128k + M.G.T.
			port_7ffd,				// If in SamRam mode, bitwise state of 74ls259.
									// For example, bit 6=1 after an OUT 31,13 (=2*6+1)
									// If in 128 mode, contains last OUT to 7ffd (paging control)
			if1paged,					// !=0 means: interface1 rom is paged in
			rldiremu,					// Bit 0: 1 if R register emulation on
									// Bit 1: 1 if LDIR emulation on
			port_fffd,					// Last OUT to fffd (soundchip register number)
			soundreg[16],				// Contents of the sound chip registers
			// ----- Z80 version 3.0 enhanced header (if PC=0 & h2len=54)
			t_l,t_m,t_h,				// T state counter
			spectator,					// Flag byte used by Spectator (QL spec. emulator)
									// Ignored by Z80 when loading, zero when saving
			MGT_ROM,				// 0xFF if MGT Rom paged
			Multiface_ROM,			// 0fFF if Multiface Rom paged. Should always be 0.
			RAM0,RAM1,				// 0xFF if 0-8191 / 8192-16383 is RAM
			joy[10],					// 5* ascii word: keys for user defined joystick
			stick[10],					// 5* keyboard mappings corresponding to keys above
        		MGT_type,				// MGT type: 0=Disciple+Epson,1=Disciple+HP,16=Plus D
			disciple_inhibit_button_status,	// 0=out, 0ff=in
			disciple_inhibit_flag;			// Disciple inhibit flag: 0=rom pageable, 0ff=not
		} head;


	PrintMsg(CStr("Loading .z80 snapshot file")); 

// read registers:
	err = File2Buffer(fileID,&head,30);
	if (err) { PrintNumberAlert(CStr(localized("Error while reading registers:","Fehler beim Lesen der Register:")),err); return err; };
	if (head.data==0x0ff) head.data = 1;

// load version 1.45 file
	if (head.pch||head.pcl)
	{	
		if (z80_hardware != zx48) HandleOptionsMenu ( i48k );
		else HandleReset (!romID);
		HandleFileMenu(iResume);

		if (head.data&0x20)		// compressed
		{	Handle scratch;

			PrintMsg(CStr("Version 1.45 compressed")); 

			scratch = TempNewHandle ( 50000, &err );
			if (err||!scratch) { PrintAlert(CStr(localized("Out of memory!","Kein Speicher mehr frei!"))); return error; }
			HLock ( scratch );

			err = File2Buffer(fileID,*scratch,50000); if ((err=eofErr)) err=noErr;
			if (err)
				PrintNumberAlert(CStr(localized("Error while reading RAM image:","Fehler beim Lesen des Speicherauszuges:")),err);
			else
			{	err = DecompressZ80 ( (Char*)*scratch, tfu, CORE+0x4000, 0x0c000 );
				if (err) PrintAlert(CStr(localized("The RAM image is corrupted","Der Speicherabzug ist defekt")));
			}
			DisposHandle ( scratch );
		}
		else			 		// uncompressed
		{	PrintMsg(CStr("Version 1.45 uncompressed"));
			err = File2Core ( fileID, 0x4000, 0x0c000 );
			if (err)  PrintNumberAlert(CStr(localized("Error while reading RAM image:","Fehler beim Lesen des Speicherauszuges:")),err);
		};
		if (err) return err; else goto xok;
	};

// read extended header, at least for version 2.01
	err = File2Buffer ( fileID, &head.h2lenl, 25 );
	if (err) { PrintNumberAlert(CStr(localized("Error while reading extended header:","Fehler beim Lesen der Header-Erweiterung:")),err); return err; }
	switch ( head.h2lenl+256*head.h2lenh )
	{	case 23:	PrintMsg(CStr("Version 2.01"));	break;		// version 2.01 header: ok
		case 54:	PrintMsg(CStr("Version 3.0"));	break;		// version 3.0 header: ok
		default:	PrintAlert(CStr(localized("Unrecognized version of .z80 file","Unbekannte Version des .z80 Dateiformats"))); return error;
	}

	head.pcl=head.npcl; head.pch=head.npch;

	if (head.if1paged) { PrintAlert(CStr(localized("Interface-1 ROM paged in (not supported)","Interface-1 ROM ist eingeblendet (nicht unterstützt)"))); return error; };		

// load vs. 3.0 header extension
	if (head.h2lenl==54)
	{	err = File2Buffer ( fileID, &head.t_l, 31 );
		if (err) { PrintNumberAlert(CStr(localized("Error while reading extended header:","Fehler beim Lesen der Header-Erweiterung:")),err); return err; }
	}

// convert header from vs. 2.0 -> 3.0
	if (head.h2lenl==23)
	{
		if (head.model>=2) head.model += 1;
		head.t_h = 3; 									// •••
		head.t_m = frame_cycles/4/256;					// •••
		head.t_l = frame_cycles%256;						// •••
		head.spectator=0;
		head.MGT_ROM = 0;
		head.Multiface_ROM = 0;
		head.RAM0 = head.RAM1 = 0;
		memset ( head.joy, 0,10 );
		memset ( head.stick, 0,10 );
		head.MGT_type = 0;
		head.disciple_inhibit_button_status = 0;
		head.disciple_inhibit_flag = 0;
		
		head.h2lenl = 54;
	}

// load version 3.0 file
	if (head.h2lenl==54)
	{	short loaded;

// select required hardware
		switch (head.model)	
		{
		case 1:	PrintMsg (CStr(localized("Warning: Interface-1 may be required!","Achtung: Interface-1 könnte benötigt werden!")));
		case 0:	PrintMsg ( CStr(localized("Model = ZX Spectrum 48k", "Modell = ZX Spectrum 48k")) );
				if (z80_hardware != zx48) HandleOptionsMenu ( i48k );
				else HandleReset (!romID);
				break;
		case 5:	PrintMsg (CStr(localized("Warning: Interface-1 may be required!","Achtung: Interface-1 könnte benötigt werden!")));
		case 4:	PrintMsg ( CStr(localized("Model = ZX Spectrum 128k", "Modell = ZX Spectrum 128k")) );
				if (z80_hardware != zx128) HandleOptionsMenu ( i128 );
				else HandleReset (!romID);
				break;
		case 6:
		case 2:	PrintAlert (CStr(localized("M.G.T. interface required (not supported)","M.G.T.-Interface erforderlich (nicht unterstützt)")));
				return error;
		case 3:	PrintAlert (CStr(localized("SAMRAM required (not supported)","SAMRAM erforderlich (nicht unterstützt)")));
				return error;
		default:	PrintNumberAlert (CStr(localized("Illegal hardware configuration:","Ungültige Hardwarenummer:")),head.model);
				return error;
		}
		HandleFileMenu(iResume);

// set T cycle counter
		TOTAL -= CYCLES;
		CYCLES = (signed short)(head.t_m*256+head.t_l) + frame_cycles/4*(3-(head.t_h+1)%4) - frame_cycles;
		TOTAL += CYCLES;

// loop: load all pages
		loaded = 0;
		while (!err)
		{	struct  { Char  lenl,lenh,nr; }  blockhead;	// page header
			Char *q; int len;
		
// read page header
			err = File2Buffer (fileID, &blockhead, 3 );
			if (err)
			{	if (loaded==0x000E) goto xok;		// are all paged in pages loaded ?
				PrintNumberAlert(CStr(localized("Error while reading page header:","Fehler beim Lesen eines Seiten-Headers:")),err); 
				return err;
			}
			
// read and decompress page
			len = blockhead.lenl+256*blockhead.lenh;
			q = (Char*)NewPtr ( len ); if (!q) { PrintAlert (CStr(localized("Out of memory","Zu wenig Speicher frei"))); return error; }
			File2Buffer ( fileID, q, len );
			if (err) { DisposePtr ( (Ptr)q ); PrintNumberAlert(CStr(localized("Error while reading page:","Fehler beim Lesen einer Seite:")),err); return err; }
			
			if (head.model<=2)	// zx spectrum 48k
			{	switch (blockhead.nr)
				{	case 0:	PrintMsg ( CStr("ID 0: The ROM (skipped)"));		break;
					case 1:	PrintMsg ( CStr("ID 1: Interface-1 ROM (skipped)"));	break;
					case 11:	PrintMsg ( CStr("ID 11: Multiface ROM (skipped)"));	break;
					case 4:	PrintMsg ( CStr("ID 4: RAM $8000-$BFFF"));	err = DecompressZ80 ( q,len,WrPtr(0x8000),0x4000 );	loaded |= 4; break;
					case 5:	PrintMsg ( CStr("ID 5: RAM $C000-$FFFF"));	err = DecompressZ80 ( q,len,WrPtr(0xc000),0x4000 );	loaded |= 8; break;
					case 8:	PrintMsg ( CStr("ID 8: RAM $4000-$7FFF"));	err = DecompressZ80 ( q,len,WrPtr(0x4000),0x4000 );	loaded |= 2; break;
					default:	PrintNumberAlert(CStr(localized("Illegal page id: ","Ungültige Seitennummer: ")), blockhead.nr); DisposPtr((Ptr)q); return error;
				};
			}
			else				// zx 128k
			{	switch (blockhead.nr)
				{	case 0:	PrintMsg(CStr("ID 0: Basic ROM (skipped)"));		break;
					case 1:	PrintMsg(CStr("ID 1: Interface-1 ROM (skipped)"));	break;
					case 2:	PrintMsg(CStr("ID 2: Boot ROM (skipped)"));		break;
					case 11:	PrintMsg(CStr("ID 11: Multiface ROM (skipped)"));	break;
					case 3:
					case 4:
					case 5:
					case 6:
					case 7:	
					case 8:
					case 9:
					case 10:	PrintMsg ( CatStr ( CatStr(CStr("ID "),NumStr(blockhead.nr)), CatStr(CStr(": RAM page "),NumStr(blockhead.nr-3)) ) );
							err = DecompressZ80 ( q, len, ram[blockhead.nr-3], 0x4000 );
							if (blockhead.nr-3==2) loaded |= 4;
							if (blockhead.nr-3==5) loaded |= 2;
							if (blockhead.nr-3==(head.port_7ffd&7)) loaded |= 8;
							break;
					default:	PrintNumberAlert(CStr(localized("Illegal page id: ","Ungültige Seitennummer: ")), blockhead.nr); DisposPtr((Ptr)q); return error;
				};
			}
			
			DisposPtr ( (Ptr)q );

		};	// loop exit if err!=ok:
		PrintAlert(CStr(localized("Compressed page image is corrupted","Komprimierte Speicherseite ist defekt"))); 
		return err;
	}

// -----	store registers and resume execution
xok:	RI 	= head.i;
	RR 	= head.r&0x7f + head.data<<7;		// bit 7 von r steckt in data bit 0 ... tzz tzz
	RA 	= head.a;		RA2 	= head.a2;		
	RF	= head.f;		RF2	= head.f2;
	RB	= head.b;		RC	= head.c;	
	RD	= head.d;		RE	= head.e;
	RH	= head.h;		RL	= head.l;
	XH	= head.xh;		XL	= head.xl;
	YH	= head.yh;		YL	= head.yl;
	SPH	= head.sph;	SPL	= head.spl;	
	PCH	= head.pch;	PCL	= head.pcl;
	BC2	= head.b2*256 + head.c2;	
	DE2	= head.d2*256 + head.e2;
	HL2	= head.h2*256 + head.l2;
	IFF1 	= head.iff1;
	IFF2	= head.iff2;
	bordercolor = (head.data/2)&0x07;
	IM = head.im&3;	if (IM>2) { PrintAlert(CStr(localized("Illegal irpt mode","Interruptmodus nicht gültig")));  return error; }

	if (z80_hardware!=zx128&&z80_hardware!=zxplus2)
	{	Set_AY_Register(7,0xff);						// disable all sound ((security only))
	}
	else
	{	int i; 
		sndlogptr = sndlog; 
		for (i=16;i--;) Set_AY_Register(i,head.soundreg[i]);	// sound chip data registers
		Select_AY_Register(head.port_fffd);				// sound chip register select register
		ZX128PageCtrl ( head.port_7ffd );					// RAM/ROM page selection
	}

	return ok;
}


// -----	Load ROM from file -------------------------------------------------
//		Load ROM from file and reset Specci
//		return ok	-> you should rename specci window
//		else		-> you should reload roms and reset specci
OSErr LoadROM ( short fileID )
{	OSErr err;

	PrintMsg(CStr("Load .ROM file")); 

	err = File2Buffer ( fileID, rom[0], 0x4000 );
	if (err)	PrintNumberAlert(CStr(localized("Error while reading file:","Fehler beim Lesen der Datei:")),err);
	else		HandleReset(false);

	romID = err?0:fileID;
	return err;
}


// -----	Save ROM to file -------------------------------------------------
OSErr SaveROM(short fileID)
{	OSErr err;

	PrintMsg(CStr("Save .ROM file")); 

	err = Buffer2File ( rom[0], fileID, 0x4000 );
	if (err) PrintNumberAlert(CStr(localized("Error while writing file:","Fehler beim Schreiben der Datei:")),err);
	return err;
}


// -----	Load ROM from Resource 'ZRom' #48 ----------------------------
OSErr LoadResROM ( int rom_id, Char* page )
{	Handle	romHandle;

	PrintDebug(CatStr(CStr("Loading ROM from resource: "),NumStr(rom_id)));
	
	romHandle = GetResource ( 'ZRom', rom_id );
	if (!romHandle) 
	{	PrintAlert ( CStr(localized(	"The ZX Spectrum ROM could not be read from the resource fork!",
								"Das ZX Spectrum ROM konnte nicht aus der Ressource geladen werden!" )));
		return error;
	}

	Buffer2Buffer ( *romHandle, page, 0x4000 );
	ReleaseResource(romHandle);
	romID=0;

// Apply patches:	
	if (rom_id==48 || rom_id==1281 || rom_id==1283 )	// ZX 48k ROM and ZX 128 Basic ROMs are very similar!
	{	page[0x0556] = 199;		// Patch LOAD/VERIFY
		page[0x04c2] = 199;		// Patch SAVE
		page[0x03D5] = 199;		// Patch BEEP
	}

	return ok;
}


// ----	Load .scr screenshot file ----------------------------------------
OSErr LoadSCR(short fileID)
{	OSErr err;

	PrintMsg(CStr("Load .SCR screenshot")); 

	err = File2Buffer ( fileID, video_page, 6912 );
	if (err) PrintNumberAlert(CStr(localized("Error reading file:","Fehler beim Lesen der Datei:")),err);
	else
	{	Char a,b,c,d; 
		a = Peek(0x4000+6144+0) & 0x78;
		b = Peek(0x4000+6144+31) & 0x78;
		c = Peek(0x4000+6144+672) & 0x78;
		d = Peek(0x4000+6144+703) & 0x78;
		if ( a==b || a==c || a==d ) {}
		else	if ( b==c || b==d ) a=b;
		else	if ( c==d ) a=c;
		bordercolor = (a>>3)&0x07;
		HandleFileMenu(iSuspend);
	}
//	return err;	
//	return err==eofErr?noErr:err;	// allow loading of arbitrary data into the screen
	return noErr;					// failed SCR load should not reset
}


// -----	Save .scr screenshot file ----------------------------------------
OSErr SaveSCR(short fileID)
{	OSErr err;

	PrintMsg(CStr("Save .SCR screenshot")); 

	err = Buffer2File ( video_page, fileID, 6912 );
	if (err) PrintNumberAlert(CStr(localized("Error writing file:","Fehler beim Schreiben der Datei:")),err);
	return err;
}


// -----	Save Screenshot in pict file -----------------------------------------
OSErr SavePICT(short fileID)
{	OSErr 	err;
	long		picLen,len;
	Char		hdr[512];
	PicHandle	picHandle	= nil;	// where QD will record Picture
	Rect		box;
	Short	x,y,i; Pattern p; Rect r;
	Char		*sattr, *sptr;

// Record pict:
	if (!zxwindow) return error;
	SetPort(zxwindow);
	ClipRect ( &zxwindow->portRect );
	SetRect ( &box, hborder, vborder, hborder+256, vborder+192 );
	picHandle = OpenPicture ( &box );
	if (!picHandle) { PrintAlert(CStr("Error allocating picture handle")); return error; };

	for ( y=0; y<192; y+=8 )
	{	sptr = video_page + ((y&0x0038)<<2) + ((y&0x00c0)<<5);
		sattr = video_page + 6144 + (y<<2);
		r.right = (r.left=hborder) +8;
		r.bottom = (r.top=vborder+y) +8;
		for ( x=0; x<32; x++ )
		{	for ( i=0; i<8; i++ )  p.pat[i] = sptr[i*256];
			RGBForeColor ( rgb+ink[*sattr] );
			RGBBackColor ( rgb+pap[*sattr] );
			FillRect ( &r, &p );
			r.left+=8; 
			r.right+=8;
			sptr++;
			sattr++;
		};
	};

	ClosePicture();
	
// Save:
	memset ( hdr, 0,len=512 );
	err = FSWrite ( fileID, &len, hdr );
	if (err) { PrintNumberAlert(CStr(localized("Error writing header:","Fehler beim Schreiben des Datenkopfes:")),err); goto xx; };

	HLock ( (Handle)picHandle );
	picLen = GetHandleSize ( (Handle)picHandle );
	err = FSWrite ( fileID, &picLen, *picHandle );
	if (err) { PrintNumberAlert(CStr(localized("Error writing picture:","Fehler beim Schreiben des Bildes:")),err); goto xx; };

xx:	KillPicture ( picHandle );
	return err;
}

