
#include	<folders.h>

#include	"kio.h"
#include	"options.h"
#include	"strings.h"


/*	Preferences handling
	this module provides means to store and retrieve preferences.


***	Copyright:	1994-1996 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.



The preferences are stored in a plain text file.
They may therefore be easily edited with a standard text editor.
Only text strings which do not contain weird characters may be stored.

Each line in the text file represents one setting. 
The layout of each line is as follows:

AAAATTTTXXXXXXXXXXXXXXXXXX...

AAAA			application signature
    TTTT		identifier of the setting
        XXXX...	text describing the setting

It is possible to store the settings of multiple applications in the same file.
Applications may access and modify settings of other applications using the same file 
though the procedures of this module do not support this.


There are four settings which must be defined in the file "options.h":

long	APPLID		the four byte signature of your application		=> use the applications creator ID
Str31	PREFNAME	the name of the preferences file				=> default file should be "kio's prefs"
long	PREFGAP		amount of spare bytes to be allocated			=> one or two thousands won't hurt
long	SCRAMBLE	application-specific scrambling byt				=> required for protected entries only 


There are four basic procedures to store and retrieve data:

OSErr OpenPrefs() reads the preferences into memory and locks them.
				The preferences file is kept locked until ClosePrefs() is called.
OSErr ClosePrefs() writes the preferences back to file, purges the memory and releases file.
OSErr PutPref (long id, string s) stores the settings s in the prefs memory block.
				The settings are marked for application APPLID with identifier id.
				This may overwrite an existing setting.
				If the memory must be increased, this is done by using the spare bytes as reserved by PREFGAP.
				Only if this is not sufficient, the prefs memory may move when it is increased.
				In this case PutPref() increases the memory and does not store the settings and returns 
				an error. You may reissue the request. Caveat: If the memory has moved, all strings which have
				been retrieved with GetPref() become invalid. 
OSErr	GetPref (long id, string *s) retrieves the settings from the prefs memory block.
				The settings for application APPLID with identifier id are retrieved.
				If they are not found, the string is left untouched and an error is returned. You may store
				the default settings for this preference in prior to calling this procedure.
				You should always validate the retrieved values before using them! Otherwise windows open 
				outside the screen or other mysterious things may happen.

Additional procedures:

OSErr	PutSecurePref (long id, string s) works like PutPref(), except that a checksum is appended.
OSErr	GetSecurePref (long id, string *s ) retrieves data stored with PutSecurePref().
OSErr	PutHexPref (long id, Char*p, long n) works like PutPref(), expcept that arbitrary binary data may be stored
OSErr	GetHexPref (long id,, Char*p, long n) retrieves data stored with PutHexPref().
*/


short	prefID	= 0;
Handle	prefHdl	= nil;
Char*	prefPtr	= nil;
long	prefLen	= 0;
OSErr	prefErr	= ok;
Boolean	prefMod	= false;


OSErr ClosePrefs()
{
	if (prefPtr&&prefID&&!prefErr&&prefMod) 
	{	prefErr=SetFPos (prefID, fsFromStart, 0);
		if (!prefErr) prefErr = FSWrite(prefID,&prefLen,prefPtr);
		if (!prefErr) prefErr = SetEOF (prefID, prefLen);
	}
	if (prefHdl) DisposeHandle(prefHdl); prefHdl = nil; prefPtr = nil;
	if (prefID)  FSClose(prefID);        prefID  = 0;
	return  prefErr;
}


OSErr OpenPrefs()
{	FSSpec	prefSpec;
	FInfo	fndrInfo;

	if (prefID) return prefErr=ok;
	prefMod = false;
	prefErr = FindFolder ( kOnSystemDisk, 'pref', kCreateFolder, &prefSpec.vRefNum, &prefSpec.parID );
	MemCopy ( PREFNAME, prefSpec.name, 32 );
	if (!prefErr) prefErr = FSpCreate ( &prefSpec, 'ttxt', 'TEXT', 0 ); if (prefErr==dupFNErr) prefErr=noErr;

	if (!prefErr) prefErr = FSpGetFInfo (&prefSpec, &fndrInfo);
	fndrInfo.fdFlags |= fInvisible;
	if (!prefErr) prefErr = FSpSetFInfo (&prefSpec, &fndrInfo);

	if (!prefErr) prefErr = FSpOpenDF ( &prefSpec, fsRdWrPerm, &prefID );
	if (!prefErr) prefErr = GetEOF ( prefID, &prefLen);
	if (!prefErr) prefErr = SetFPos ( prefID, fsFromStart, 0 );
	if (!prefErr) prefHdl = TempNewHandle (prefLen+PREFGAP, &prefErr); 
	if (!prefErr) { HLock(prefHdl); prefPtr = (Char*)*prefHdl; }
	if (!prefErr) prefErr = FSRead ( prefID, &prefLen, prefPtr);
	if (prefErr) ClosePrefs();	// won't save
	return prefErr;
}


OSErr GetPref(long id, string *s)
{	Char *p,*e;

	if (!prefID) return error;

	p=prefPtr;
	e=p+prefLen;

	while (p<e)
	{
		if (*(long*)p == APPLID)	// test for right application and pref's id
		{	p+=4;
			if (*(long*)p == id )
			{	p+=4;
				s->text = p;				// return pref's string
				while (*p!=10&&*p!=13) p++;
				s->len = p-s->text;
				return ok;
			}
		}
		while (*p!=10&&*p!=13) p++;		// else skip to next line
		while (*p==10||*p==13) p++;
	}
	return error;
}


OSErr PutPref(long id, string s)
{	string d;
	long *p;

	if (!prefID) return error;

	prefMod = true;
	
	prefErr = GetPref(id,&d);			// overwrite existing id ?
	if (!prefErr)
	{
		if (d.len==s.len) { MemCopy(s.text,d.text,s.len); return ok; }			// => quick overwrite & exit
		while (d.text[d.len]==10 || d.text[d.len]==13) d.len++;					// include line break	
		if (d.text+d.len>prefPtr+prefLen) 										
			prefLen = d.text-8-prefPtr;											// overwrite last line
		else
		{	MemMove ( d.text+d.len, d.text-8, prefPtr+prefLen-(d.text+d.len) );	// => discard old setting
			prefLen -= d.len+8;
		}
	}
	
	if (prefLen+s.len+9 > GetHandleSize(prefHdl))	// increase memory ?
	{
		HUnlock(prefHdl);													
		SetHandleSize(prefHdl,prefLen+PREFGAP);
		prefErr=MemError(); if (prefErr) { ClosePrefs(); return prefErr; }	// could not increase => panic!
		HLock(prefHdl);
		prefPtr=(Char*)*prefHdl;
		return error;								// prefs memory may have moved => error
	}

	p = (long*)(prefPtr+prefLen);
	*p++ = APPLID;
	*p++ = id;
	MemCopy ( s.text, p, s.len );
	prefLen += s.len+9;
	prefPtr[prefLen-1] = 13;
	
	return prefErr=ok;
}


long Check ( string s );
long Check ( string s )
{	long crc = SCRAMBLE;
	Char *p,c;

	for ( p=s.text+s.len; --p>=s.text; ) for (c=*p;c;c>>=1) crc += crc + (crc>=0) + c;
	return crc;
}


// store preferences with a checksum
OSErr PutSecurePref ( long id, string s )
{
	return PutPref ( id, CatStr(HexStr(Check(s),8),s) );
}


// retrieve preferences and check checksum
OSErr GetSecurePref (long id, string *s)
{	string p;
	OSErr err;

	err=GetPref(id,&p); if (err) return err;
	if (!SameStr(HexStr(Check(*s),8),QSubStr(p,0,8))) return error;
	*s = QSubStr(p,8,99999);
	return ok;
}


stuff_3 ( Char *q, Char *z );
stuff_3 ( Char *q, Char *z )
{	short i; Char c;

	c = 0x01;
	for (i=3;i--;)	
	{	c = (c<<2) + (*q>>6);
		*++z = (*q++ & 0x3f) + 0x40;
	}
	z[-3] = c;
}


unstuff_3 ( Char *q, Char *z );
unstuff_3 ( Char *q, Char *z )
{	short i; Char c;

	c = *q++ - 0x40;
	for (i=3;i--;)
	{	c <<= 2;
		*z++ = (c&0xc0) + *q++ - 0x40;
	}
}


// store binary data into preferences ----------------------
OSErr PutHexPref ( long id, Char *p, long n )
{	string s;
	Char *z;
	OSErr err;

 	s = SpaceStr ( n + (n+2)/3 );
	z = s.text;

	while (n>=3)
	{
		stuff_3 ( p, z );
		p+=3;
		z+=4;
		n-=3;
	}
	
	if (n>0)
	{	Char c4[4];
		stuff_3 ( p, c4 );
		MemCopy ( c4, z, n+1 );
	}
	
	err = PutPref ( id, s );
	s.text=nil;
	return err;
}


// retrieve binary data from preferences --------------------
OSErr GetHexPref (long id, Char *p, long n)
{	string s;
	Char *q;
	OSErr err;
	
	err = GetPref(id,&s); if (err) return err;
	if (s.len != n + (n+2)/3) return error;	// struct size error
	q = s.text;
	
	while (n>=3)
	{
		unstuff_3 ( q, p );
		q+=4;
		p+=3;
		n-=3;
	}

	if (n>0)
	{	Char c3[3];
		unstuff_3 ( q, c3 );
		MemCopy ( p, c3, n );
	}
	
	return ok;
}



