
/*	Sound output sampler

***	Copyright:	1995-1999 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	24.Mar.95:	First revision										KIO !
	05.Jun.95:	Combined ear and mic sound output						KIO !
	29.Nov.95:	AY-3-8912 emulation								KIO !
	28.Dec.99:	Rework to work with MacOS 8.6 (8.5? (8.1?))	fuckin' hell!	kio
    20.Oct.01   revised for CW7Pro and ANSI C compliance									hdusel@bnv-gz.de
*/

#include	<sound.h>
#include	<memory.h>

#include	"kio.h"
#include	"z80.h"
#include	"sound.h"
#include	"application.h"
#include	"messages.h"
#include	"menue.h"
#include	"z80 main.h"
#include	"preferences.h"
#include	"terminal.h"
#include	"strings.h"
#include	"display.h"


// adjustable values:
#define	sndlogsize			5000			// number of entries in ear/mic/AY output log buffer;	suggested: 4000
#define	maxsounds		8				// number of entries in sound channel;				suggested: 6
#define	maxsamples		10000			// max. sample length;							suggested: 10000
#define	max_volume		90				// best (no distortion) = 43; loudest = 127; 			suggested: 90

// constants:
#define	ticks_per_sample_44	79			// 44 kHz: 79;  69994 T/frame
#define	sample_rate_44		0xac440000	// ((sound.h))
#define	ticks_per_sample_22	156			// 22 kHz: 157;  70022 T/frame
#define	sample_rate_22		0x56EE8BA3	// ((sound.h))

// sampling rate setting:
long		sample_rate		= sample_rate_44;		// constant for sound header
long		ticks_per_sample	= ticks_per_sample_44;	// step width in T cycles

// volume setting:
short	sound_volume		= 0;				// volume: 0 ... 7
short	sample_volume		= 0;				// volume: 0 ... 127

// sound log
long		*sndlog			= nil;			// output log buffer
long		*sndlogptr		= nil;			// output log write pointer
long		*sndlogend		= nil;			// output log end pointer

// sample buffer:
SndChannelPtr	sound_channel	= nil;			// pointer to sound channel
SoundHeader	sound_header[maxsounds];		// sound headers to be put into sound_channel
SndCommand	sound_command;				// command to put header into channel
char*		sound_buffer	= nil;			// buffer for sample buffers

// ear/mic output state
Boolean	ear_out			= 1;

// values remembered between samples
long		lastcycle			= 0;				// T cycle of last sample
short	lastsample		= 0x80;			// value of last sample (actual state of Mac's speaker)


// ====	AY-3-8912 sound chip ============================================================

#define	AYCLK	16			// AYCLK  =  8*(z80_clock/AY_clock)
short	AYregnr;				// selected register
Char		AYreg[16];			// 16 registers/ports of the AY chip
char		logvol[16];			// logarithmic volume table (( calculated by Set_AY_Volume() ))



// -----	internal representation of the sound chip's state -----------------------------------------

/*	envelope generator:
	there is 1 envelope generator. envelope shape is controlled by e_attack, e_invert and e_hold.
	the 'repeat' bit (D3 of reg. 13) is not used and always treated as set:
	therefore %00xx is converted to %1001 and %01xx is converted to %1111 which give the same results.

	D3 = repeat
	D2 = attack
	D1 = toggle
	D0 = hold

	1111, 01xx:	/_____	 repeat &  attack &  toggle &  hold		||	!repeat &  attack
	1110:		/\/\/\	 repeat &  attack &  toggle & !hold
	1101:		/~~~~~	 repeat &  attack & !toggle &  hold
	1100:		//////	 repeat &  attack & !toggle & !hold
	1011:		\~~~~~	 repeat & !attack &  toggle &  hold
	1010:		\/\/\/	 repeat & !attack &  toggle & !hold
	1001, 00xx:	\_____	 repeat & !attack & !toggle &  hold		||	!repeat & !attack
	1000:		\\\\\\	 repeat & !attack & !toggle & !hold

	the envelope generator is running free and only retriggered if a new envelope is selected.
*/	
long		e_cnt;			// next tick to inc/dec volume
long		e_reload;			// reload value for e_cnt
short	e_idx;			// volume index 0 (off) ... 15 (max)
Boolean	e_repeat;			// repeat ramping
Boolean	e_invert;			// toggle ramping direction
short	e_ramp;			// ramping up/down/off


/*	noise (all channels)
*/
#define	n_max	5000
long		n_cnt;			// next tick to toggle output(s)
short	n_reload;			// reload value for n_cnt (multiplied with n_dur[n_idx++])
char		n_dur[n_max];		// pseudo-random reload value table
short	n_idx;			// index inside n_bit[]

/*	sound channels
*/
long		s_cnt[3];			// period counter (( T cycle of expiring ))
long		s_reload[3];		// reload value for period counter
Boolean	s_env[3];			// use envelope
Boolean	s_sound_on[3];		// sound output enable
Boolean	s_noise_on[3];		// noise output enable
char		s_vol[3];			// volume of channel: 0...127 (if output on)  resp.  0...-127 (if output off)

char		channel[3];		// channel A, B, C on/off
char		ay_emulation;


// -----	Utilities ---------------------------------

#define	bit(B,N)	(((N)>>(B))&1)


// ==================================================================================


// -----	Select sampling frequency ------------------------------
void SetSampleRate ( long n )
{
	SuspendSound();
	
	if (n==sample_rate_44)	// 44 kHz
	{
		sample_rate		= sample_rate_44;
		ticks_per_sample	= ticks_per_sample_44;
	}
	else					// 22 kHz
	{
		sample_rate		= sample_rate_22;
		ticks_per_sample	= ticks_per_sample_22;
	}

	ResumeSound();

	CheckItem ( soundmenu, i22, sample_rate==sample_rate_22 );	
	CheckItem ( soundmenu, i44, sample_rate==sample_rate_44 );	
}


// -----	Get current sampling frequency (for the sake of modularity) -----------
long GetSampleRate()
{	
	return sample_rate;
}


// -----	Set sound volume --------------------------------------
void SetSoundVolume ( short n )
{
	SuspendSound();
	sound_volume = minmax ( 0,n,7 );
	ResumeSound();
	
	for ( n=0; n<=7; n++ )  CheckItem ( soundmenu, iSoundOff+n, n==sound_volume );
}


// -----	Get current sound volume (for the sake of modularity) -----------
short GetSoundVolume()
{	
	return sound_volume;
}


// -----	AY emulation on/off ---------------------------------------------
//		may be called with flags=on or flags=off  to switch on/off all channels
//		may be called with flags=channel_A+channel_B+channel_C to switch on/off channels individually
void SetSoundChannels ( char flags )
{	
	SuspendSound();
	if (flags==1) flags = channel_A+channel_B+channel_C;
	ay_emulation = 
		( channel[0] = flags&channel_A) +
		( channel[1] = flags&channel_B) +
		( channel[2] = flags&channel_C);
	ResumeSound();	

	CheckItem ( soundmenu, iAY, ay_emulation );
	CheckItem ( soundmenu, iAY_A, channel[0] );
	CheckItem ( soundmenu, iAY_B, channel[1] );
	CheckItem ( soundmenu, iAY_C, channel[2] );
}


// -----	Get current emulation state (for the sake of modularity) -----------
char GetSoundChannels(void)
{	
	return channel[0] + channel[1] + channel[2];
}


// -----	Suspend sound generation when application is juggled out
//		stop sound immediately
//		dispose all channels and memory
void SuspendSound (void)
{
	if (sound_channel)  SndDisposeChannel ( sound_channel, true );	// dispose&quiet immediately
	sound_channel = nil;

	if (sndlog)  DisposPtr ( (Ptr)sndlog );
	sndlog = sndlogptr = sndlogend = nil;

	if (sound_buffer) DisposPtr ( sound_buffer );
	sound_buffer = nil;
}


// -----	Resume sound generation when application is juggled in
//		This function completely initializes the sound module
//		according to sound_method and sound_volume
void ResumeSound (void)
{	OSErr err;  char c; short i; long n;

	if (sound_channel) return; 		// not suspended
	sound_volume = minmax(0,sound_volume,7);
	if (!sound_volume)  return;		// sound off

// allocate log buffer
	sndlog = (long*)NewPtr ( (sndlogsize+1)*2*sizeof(long) );
	if (!sndlog) goto xx;
	sndlogend = sndlog + 2*sndlogsize;
	sndlogptr = sndlog;

// allocate sound channel
	err = SndNewChannel (&sound_channel, sampledSynth, initMono+initNoInterp, nil);
	if (err||!sound_channel) { sound_channel=nil; goto xx; }
	sound_channel->qLength = maxsounds-1-1;	// reduce queue length; caveat: the queue can hold qLength+1 commands !!!

// allocate sample buffer
	sound_buffer = NewPtr ( maxsounds*maxsamples );
	if (!sound_buffer) goto xx;

// initialize sound queue chunks
	sound_header->samplePtr		= sound_buffer;
	sound_header->length		= 0;
	sound_header->sampleRate	= sample_rate;
	sound_header->loopStart		= 0;
	sound_header->loopEnd		= 0;
	sound_header->encode		= stdSH;			// standard sound header
	sound_header->baseFrequency	= 60;			// C

	for ( i=1; i<maxsounds; i++ )
	{	sound_header[i] = sound_header[i-1];
		sound_header[i].samplePtr += maxsamples;
	}

	sound_command.cmd			= bufferCmd;
	sound_command.param1		= 0;
	sound_command.param2		= (long)sound_header;

// calculate volume
	n = max_volume*100;
	for ( i=7-sound_volume; i--; ) n = n*75/100;
	sample_volume = n/100;

	lastcycle			= TOTAL-CYCLES;	// now
	lastsample		= 0x0080;		// ±0
	ear_out			= 1;				// high


	if (!ay_emulation) return;


// calculate volume table for AY sound chip
	n = sample_volume*100;
	for ( i=16; i--; )
	{	logvol[i] = n/100;
		n = n * 82 / 100;
	};
	logvol[0] = 0;

// calculate noise table
	GetDateTime((unsigned long*)&qd.randSeed);
	for ( i=n_max; i--; )
	{	for (n=1;n<16;n++) if (Random()&1) break;
		n_dur[i] = n;
	};

// recalculate state of AY sound engine
	n_reload =   ((AYreg[6]&0x1f)+0x20) * AYCLK;
	n_cnt = TOTAL-CYCLES;
	n_idx = n_max;

	e_reload = ( AYreg[11] + 256*AYreg[12] ) * AYCLK * 2;
	c=AYreg[13];
	if (!(c&8)) c = c&4 ? 0x0f : 0x09;
	e_idx      = bit(2,c) ? 0 : 15;
	e_ramp   = bit(2,c) ? 1 : -1; 
	e_invert = bit(1,c); 
	e_repeat = !bit(0,c);
	e_cnt = TOTAL-CYCLES;

	c=~AYreg[7];
	for (n=3;n--;)
	{	s_reload[n] = ( AYreg[n*2] + 256*(AYreg[n*2+1]&0x0f) ) * AYCLK;	
		s_sound_on[n] = bit(n,c);
		s_noise_on[n] = bit(n+3,c);
		s_cnt[n] = TOTAL-CYCLES;									
		s_env[n] = bit(4,AYreg[n+8]);									
		s_vol[n] = logvol [ s_env[n] ? e_idx : AYreg[n+8]&0x0f ];
		if (!s_sound_on[n]) s_vol[n] *= -1;
	}
	
	lastsample  =  0x0080  +  (s_vol[0]>0?s_vol[0]:0)  +  (s_vol[1]>0?s_vol[1]:0)  +  (s_vol[2]>0?s_vol[2]:0)  /* -  (!ear_out)*sample_volume */;
	
	if (debug>=log_level)
	{	PrintStr(CStr("AY-3-8912 volume table:"));	
		for (i=0;i<=15;i++)
		{	PrintStr(NumStr(logvol[i]));
		}
		PrintNL();
	}


	return;


// error	
xx:	SuspendSound();		// dispose off buffers
	PrintAlert ( CStr(localized("Not enough memory for sound output!", "Nicht genug Speicher für die Tonausgabe!" )));
	SetSoundVolume ( off );
}


/* -----	Add one frame of sampled sound ----------------------------------------------------------
*/
void PlaySample (void)
{	register long cycle,tps,*q;   register short sam; register char *z;
	long n,now;   OSErr err; SoundHeader *shp; Char c;

//	if (!sndlog) return;					// security
//	if (!sound_volume)  return;			// sound off


// synchronize virtual time to real time:
/* note:
	once i thought i could use approx. 4 sound commands with associated sample buffers and feed them to the sound device
	at soon it is ready to accept another buffer. i thought, this would ideally result in always 3 buffers in the sound device and 
	one buffer currently under construction, giving me a time buffer corresponding to the 3 buffers already in the device, 
	e.g. for not-so-cooperative background activities or longish screen updates. 
	this worked fine up to MacOS 8.0 ... 8.5. i don't know when exactly things changed.
	
	now, with MacOS 8.6 on my PowerMac G4 things were bad.
	it seems, after a day of investigation, that my buffers are fed into another, larger buffer of the sound device itself.
	so the advantage of my 4 small buffers are void. if a sound device buffer becomes empty, i can feed for several frames
	sound without delay. when the sound device's buffer becomes full, i have to wait. then again, i can feed several frames without
	interruption. to make things worse, i can no longer detect a buffer-full state by comparing the head and tail pointers of the sound
	queue. i can only tell that the queue is full, after i have unsuccessfully added a sample buffer or i can tell that it underrun
	(with audible click). i can no longer implement a low/high water mark algorithm and react to low water state by skipping a
	screen refresh before the water is out.

	for mac spectacle i try to fix it like this:
	1. i synchronize to the ticker interrupt as if no sound is playing: frames_done++
	2. if i detect underrun i set hurry_up (-> skipping one screen refresh) and adjust frames_done := frames-4
	3. if i detect overrun i adjust frames_done := frames

	additional note:
	keep in mind that frames and frames_done may wrap from MAX_VALUE to MIN_VALUE.
	so never compare e.g. (frames_done<frames) but always (frames_done-frames>0)
*/

// ---1---
	hurry_up = false;
	frames_done++;

// ---2---
	if (!sndlog || !sound_volume || run_unlimited)
	{
		if (frames_done-frames+4<0) frames_done=frames-4;
		while (frames_done-frames>0) if (mbar_present) PollEvents(0,0,nil);
		if (frames_done<frames) hurry_up = true;
		return;
	}
	else if (sound_channel->qHead==-1 || frames_done-frames+4<0) 
	{ 
		frames_done=frames-4; 
		hurry_up=true; 
	}


	cycle = lastcycle;					// sampling start
	now	= TOTAL-CYCLES;				// sampling end
	tps = ticks_per_sample;				// sampling frequency

// calculate sample size, check for zero length sample (e.g. after LDIR), and too long samples
	n = ( now - lastcycle + tps - 1 ) / tps;
	if ( n<=1 )  return;
	if (n>=maxsamples) cycle = now - tps*(maxsamples-1);

// add final sampling point
	*sndlogptr = now;

// get free buffer
	shp = (SoundHeader*)sound_command.param2;
	if ( (++shp) == (sound_header+maxsounds) ) shp = sound_header;
	sound_command.param2 = (long)shp;

// last preparations
	q = sndlog;						// source values
	z = shp->samplePtr;					// target pointer
	sam  = lastsample;					// actual sample byte


	if ( ay_emulation )	// ------------->	sample sound: ear/mic/AY
	{
		while ( true )
		{
			while ( (*q-cycle)>0 ) 
			{
				if ( e_ramp &&e_reload && (e_cnt-cycle)<=0 )	// envelope tick
				{	
					e_cnt += e_reload;
					
					e_idx += e_ramp;
					
					if (e_idx&0xF0)			// ramp finished
					{	
						if (e_repeat)
						{	if (e_invert) { e_idx = ~e_idx; e_ramp = -e_ramp; }
						}
						else
						{	e_ramp=0; if (!e_invert) e_idx = ~e_idx;
						}
						e_idx  &= 0x0F;
					}
	
	#define	chg_vol(V,N)		{ if (V>0) { sam += N-V; V = N; } else V = -N; }
	#define	ramp_channel(N)	if (s_env[N]) chg_vol(s_vol[N],logvol[e_idx])
					ramp_channel(0);
					ramp_channel(1);
					ramp_channel(2);
				}
		
		
				if ( (n_cnt-cycle)<=0 )			// noise tick
				{	
					n_cnt += n_reload * n_dur[--n_idx]; if (!n_idx) n_idx = n_max;
					
	#define	toggle(V)			sam += (V=-V)
	#define	noise_channel(N)	if (s_noise_on[N] && channel[N]) toggle(s_vol[N])
					noise_channel(0);
					noise_channel(1);
					noise_channel(2);
				}
	
	#define	sound_channel(N)	if ( s_sound_on[N] && channel[N] && s_reload[N] && (s_cnt[N]-cycle)<=0 )		\
							{ s_cnt[N] += s_reload[N]; toggle(s_vol[N]); }
	
				sound_channel(0);
				sound_channel(1);
				sound_channel(2);
	
				cycle += tps;		// add one sample
				*z++ = sam<=255?sam:255; 
			}
		
			if (q==sndlogptr) break;	// finished !
	
	#if AYCLK == 16
	#define	reload_lo(R)		R = (R&0x000FF000) + 0x0010 * c
	#define	reload_hi(R)		R = (R&0x00000FF0) + 0x1000 * (c&0x0F)
	#define	reload_low(R)		R = (R&0x001FE000) + 0x0020 * c
	#define	reload_high(R)		R = (R&0x00001FE0) + 0x2000 * c
	#else
	#define	reload_lo(R)		R = AYCLK * ( ((R/AYCLK)&0xFF00) + c )
	#define	reload_hi(R)		R = AYCLK * ( ((R/AYCLK)&0x00FF) + (c&0x0F)*0x100 )
	#define	reload_low(R)		R = AYCLK*2 * ( ((R/(AYCLK*2))&0xFF00) + c )
	#define	reload_high(R)		R = AYCLK*2 * ( ((R/(AYCLK*2))&0x00FF) + c*0x100 )
	#endif

			q++;
			c=*q;				// value written to register
	
			switch(*q++>>8)		// register number
			{
	#define	pitch_fine(N)												\
				reload_lo(s_reload[N]); if ((s_cnt[N]-cycle)<0) s_cnt[N] = cycle/*+s_reload[N]*/;	\
				if (s_reload[N]||s_vol[N]>0||!s_sound_on[N]) continue; 			\
				toggle(s_vol[N]); continue;
	
	#define	pitch_coarse(N)											\
				reload_hi(s_reload[N]); if ((s_cnt[N]-cycle)<0) s_cnt[N] = cycle/*+s_reload[N]*/;	\
				if (s_reload[N]||s_vol[N]>0||!s_sound_on[N]) continue;	 		\
				toggle(s_vol[N]); continue;
	
			case 0:	pitch_fine(0);		// channel A fine
			case 1:	pitch_coarse(0);	// channel A coarse
			case 2:	pitch_fine(1);		// channel B fine
			case 3:	pitch_coarse(1);	// channel B coarse
			case 4:	pitch_fine(2);		// channel C fine
			case 5:	pitch_coarse(2);	// channel C coarse
	
			case 6:	// noise period
				n_reload = (0x20+(c&0x1F))*AYCLK;
				continue;
	
			case 11:	// envelope fine
				reload_low(e_reload); if ((e_cnt-cycle)<0) e_cnt = cycle;
				continue;
			case 12:	// envelope coarse
				reload_high(e_reload); if ((e_cnt-cycle<0)) e_cnt = cycle;
				continue;
	
			case 7:	// mixer
				c=~c;
	#define	switch_channel(N)	s_noise_on[N]=bit(3+N,c); if (s_sound_on[N]!=bit(N,c)) 				\
							{ if ((s_sound_on[N]=!s_sound_on[N])^(s_vol[N]>0)) toggle(s_vol[N]); if ((s_cnt[N]-cycle)<0) s_cnt[N] = cycle/*+s_reload[N]*/; }
				switch_channel(0);
				switch_channel(1);
				switch_channel(2);
				continue;
	
	#define	set_volume(N)		c=logvol [ (s_env[N]=bit(4,c)) ? e_idx : c&0x0F ]; chg_vol ( s_vol[N], c )
	
			case 8:	set_volume(0); continue;	// channel A volume
			case 9:	set_volume(1); continue;	// channel B volume
			case 10:	set_volume(2); continue;	// channel C volume
	
			case 13:		
				if (!(c&8)) c = c&4 ? 0x0f : 0x09;								// envelope shape
				e_cnt      = cycle + e_reload;
				e_idx      = bit(2,c) ? 0 : 15;
				e_ramp   = bit(2,c) ? 1 : -1; 
				e_invert = bit(1,c); 
				e_repeat = !bit(0,c);
				ramp_channel(0);
				ramp_channel(1);
				ramp_channel(2);
				continue;
				
			case 14:	continue;	// port A
			case 15:	continue;	// port B
			
			case 16:	// sound output to ear/mic
				if ( (ear_out = !ear_out) != 0 )	sam += sample_volume;
				else  				sam -= sample_volume;					// toggle
				continue;
			}
		}
	}

	else				// ------------->	sample sound: ear/mic only

	{
		while (q<=sndlogptr)
		{	
			while ( (*q-cycle)>0 )
			{	*z++ = sam;
				cycle += tps;
			};
			sam = sample_volume-sam;	// flip 0x80 <-> 0x80+volume
			q+=2;
		};
		sam = sample_volume-sam;		// undo last flip		
	}
		

	lastsample= sam;		// remember
	lastcycle	= cycle;		// remember
	sndlogptr	= sndlog;		// reset log ptr
	

// and at last: play it!
	shp->length = z - shp->samplePtr;	
	err = SndDoCommand ( sound_channel, &sound_command, true );
	if (err==queueFull)
	{	
// ---3---
		while(err==queueFull) err = SndDoCommand ( sound_channel, &sound_command, true );
		frames_done=frames;
		if (hurry_up) do_abort("\pinternal error: PlaySample()",0,0);
	}
	if (err) PrintAlert ( CStr("PlaySample(): SndDoCommand() error") );

	if (frames_done-frames>0) frames_done=frames;
}





